import 'dart:math';
import 'dart:typed_data';

import 'package:flutter/material.dart';
import 'package:flutter_smartlook/wireframe/element_data.dart';

class WireframeUtils {
  static List<Skeleton>? skeletonsFromBorderRect(
    Rect elementRect,
    Color color,
    double width,
  ) {
    if (width.toInt() == 0) {
      return null;
    }

    return [
      Skeleton(
        rect: Rect.fromLTWH(
          elementRect.left,
          elementRect.top,
          elementRect.width,
          width,
        ),
        color: color,
      ),
      Skeleton(
        rect: Rect.fromLTWH(
          elementRect.left,
          elementRect.bottom,
          elementRect.width,
          width,
        ),
        color: color,
      ),
      Skeleton(
        rect: Rect.fromLTWH(
          elementRect.left,
          elementRect.top,
          width,
          elementRect.height,
        ),
        color: color,
      ),
      Skeleton(
        rect: Rect.fromLTWH(
          elementRect.right,
          elementRect.top,
          width,
          elementRect.height,
        ),
        color: color,
      ),
    ];
  }

  static Color createColorFromByteData(
    ByteData byteData,
    int height,
    int width,
  ) {
    int redBucket = 0;
    int greenBucket = 0;
    int blueBucket = 0;
    int pixelCount = 0;

    const sampleSize = 1;

    for (int y = 0; y < height; y += sampleSize) {
      for (int x = 0; x < width; x += sampleSize) {
        final int byteOffset = (y * width + x) * 4;
        final int argbInt = byteData.getInt32(byteOffset);

        final int r = (argbInt >> 32) & 0xFF;
        final int g = (argbInt >> 16) & 0xFF;
        final int b = (argbInt >> 8) & 0xFF;

        redBucket += r;
        greenBucket += g;
        blueBucket += b;
        pixelCount++;
      }
    }

    final avgRed = redBucket ~/ pixelCount;
    final avgGreen = greenBucket ~/ pixelCount;
    final avgBlue = blueBucket ~/ pixelCount;

    return Color.fromARGB(255, avgRed, avgGreen, avgBlue);
  }

  static Color blendColorsDirectly(Color baseColor, Color blendColor, BlendMode blendMode) {
    switch (blendMode) {
      case BlendMode.clear:
        return Colors.transparent;
      case BlendMode.src:
        return blendColor;
      case BlendMode.dst:
        return blendColor;
      case BlendMode.srcOver:
        return blendSrcOver(baseColor, blendColor);
      case BlendMode.dstOver:
        return blendDstOver(baseColor, blendColor);
      case BlendMode.srcIn:
        return Color.fromRGBO(blendColor.red, blendColor.green, blendColor.blue, baseColor.opacity * blendColor.opacity);
      case BlendMode.dstIn:
        return Color.fromRGBO(baseColor.red, baseColor.green, baseColor.blue, baseColor.opacity * blendColor.opacity);
      case BlendMode.srcOut:
        return Color.fromRGBO(blendColor.red, blendColor.green, blendColor.blue, baseColor.opacity * (1 - baseColor.opacity));
      case BlendMode.dstOut:
        return Color.fromRGBO(baseColor.red, baseColor.green, baseColor.blue, baseColor.opacity * (1 - blendColor.opacity));
      case BlendMode.srcATop:
        return blendSrcATop(baseColor, blendColor);
      case BlendMode.dstATop:
        return blendDstATop(baseColor, blendColor);
      case BlendMode.xor:
        return blendXor(baseColor, blendColor);
      case BlendMode.plus:
        return Color.fromRGBO(
          min(255, baseColor.red + blendColor.red),
          min(255, baseColor.green + blendColor.green),
          min(255, baseColor.blue + blendColor.blue),
          1,
        );
      case BlendMode.modulate:
        return blendMultiply(baseColor, blendColor);
      case BlendMode.screen:
        return blendScreen(baseColor, blendColor);
      case BlendMode.overlay:
        return blendOverlay(baseColor, blendColor);
      case BlendMode.darken:
        return blendDarken(baseColor, blendColor);
      case BlendMode.lighten:
        return blendLighten(baseColor, blendColor);
      case BlendMode.colorDodge:
        return blendColorDodge(baseColor, blendColor);
      case BlendMode.colorBurn:
        return blendColorBurn(baseColor, blendColor);
      case BlendMode.hardLight:
        return blendHardLight(baseColor, blendColor);
      case BlendMode.softLight:
        return blendSoftLight(baseColor, blendColor);
      case BlendMode.difference:
        return blendDifference(baseColor, blendColor);
      case BlendMode.exclusion:
        return blendExclusion(baseColor, blendColor);
      case BlendMode.multiply:
        return blendMultiply(baseColor, blendColor);
      case BlendMode.hue:
        return blendHue(baseColor, blendColor);
      case BlendMode.saturation:
        return blendSaturation(baseColor, blendColor);
      case BlendMode.color:
        return blendTheColor(baseColor, blendColor);
      case BlendMode.luminosity:
        return blendLuminosity(baseColor, blendColor);

      default:
        return baseColor;
    }
  }

  static Color blendSrcOver(Color baseColor, Color blendColor) {
    final alpha = blendColor.opacity + baseColor.opacity * (1 - blendColor.opacity);
    final red = ((blendColor.red * blendColor.opacity) + (baseColor.red * baseColor.opacity * (1 - blendColor.opacity))) ~/ alpha;
    final green = ((blendColor.green * blendColor.opacity) + (baseColor.green * baseColor.opacity * (1 - blendColor.opacity))) ~/ alpha;
    final blue = ((blendColor.blue * blendColor.opacity) + (baseColor.blue * baseColor.opacity * (1 - blendColor.opacity))) ~/ alpha;

    return Color.fromRGBO(red, green, blue, alpha);
  }

  static Color blendDstOver(Color baseColor, Color blendColor) {
    final alpha = baseColor.opacity + blendColor.opacity * (1 - baseColor.opacity);
    final red = ((baseColor.red * baseColor.opacity) + (blendColor.red * blendColor.opacity * (1 - baseColor.opacity))) ~/ alpha;
    final green = ((baseColor.green * baseColor.opacity) + (blendColor.green * blendColor.opacity * (1 - baseColor.opacity))) ~/ alpha;
    final blue = ((baseColor.blue * baseColor.opacity) + (blendColor.blue * blendColor.opacity * (1 - baseColor.opacity))) ~/ alpha;

    return Color.fromRGBO(red, green, blue, alpha);
  }

  static Color blendSrcATop(Color baseColor, Color blendColor) {
    final alpha = baseColor.opacity;
    final red = (blendColor.red * blendColor.opacity * baseColor.opacity + baseColor.red * (1 - blendColor.opacity)) ~/ alpha;
    final green = (blendColor.green * blendColor.opacity * baseColor.opacity + baseColor.green * (1 - blendColor.opacity)) ~/ alpha;
    final blue = (blendColor.blue * blendColor.opacity * baseColor.opacity + baseColor.blue * (1 - blendColor.opacity)) ~/ alpha;

    return Color.fromRGBO(red, green, blue, alpha);
  }

  static Color blendDstATop(Color baseColor, Color blendColor) {
    final alpha = blendColor.opacity;
    final red = (baseColor.red * baseColor.opacity * blendColor.opacity + blendColor.red * (1 - baseColor.opacity)) ~/ alpha;
    final green = (baseColor.green * baseColor.opacity * blendColor.opacity + blendColor.green * (1 - baseColor.opacity)) ~/ alpha;
    final blue = (baseColor.blue * baseColor.opacity * blendColor.opacity + blendColor.blue * (1 - baseColor.opacity)) ~/ alpha;

    return Color.fromRGBO(red, green, blue, alpha);
  }

  static Color blendXor(Color baseColor, Color blendColor) {
    final alpha = baseColor.opacity + blendColor.opacity - 2 * baseColor.opacity * blendColor.opacity;
    final red = (baseColor.red * baseColor.opacity * (1 - blendColor.opacity) + blendColor.red * blendColor.opacity * (1 - baseColor.opacity)).toInt();
    final green = (baseColor.green * baseColor.opacity * (1 - blendColor.opacity) + blendColor.green * blendColor.opacity * (1 - baseColor.opacity)).toInt();
    final blue = (baseColor.blue * baseColor.opacity * (1 - blendColor.opacity) + blendColor.blue * blendColor.opacity * (1 - baseColor.opacity)).toInt();

    return Color.fromRGBO(red, green, blue, alpha);
  }

  static Color blendMultiply(Color baseColor, Color blendColor) {
    return Color.fromRGBO(
      (baseColor.red * blendColor.red) ~/ 255,
      (baseColor.green * blendColor.green) ~/ 255,
      (baseColor.blue * blendColor.blue) ~/ 255,
      1,
    );
  }

  static Color blendScreen(Color baseColor, Color blendColor) {
    return Color.fromRGBO(
      255 - (((255 - baseColor.red) * (255 - blendColor.red)) ~/ 255),
      255 - (((255 - baseColor.green) * (255 - blendColor.green)) ~/ 255),
      255 - (((255 - baseColor.blue) * (255 - blendColor.blue)) ~/ 255),
      1,
    );
  }

  static Color blendOverlay(Color baseColor, Color blendColor) {
    int overlayChannel(int baseC, int blendC) {
      return baseC < 128 ? 2 * baseC * blendC ~/ 255 : 255 - 2 * (255 - baseC) * (255 - blendC) ~/ 255;
    }

    return Color.fromRGBO(
      overlayChannel(baseColor.red, blendColor.red),
      overlayChannel(baseColor.green, blendColor.green),
      overlayChannel(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendDarken(Color baseColor, Color blendColor) {
    return Color.fromRGBO(
      min(baseColor.red, blendColor.red),
      min(baseColor.green, blendColor.green),
      min(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendLighten(Color baseColor, Color blendColor) {
    return Color.fromRGBO(
      max(baseColor.red, blendColor.red),
      max(baseColor.green, blendColor.green),
      max(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendColorDodge(Color baseColor, Color blendColor) {
    int colorDodgeChannel(int c1, int c2) {
      return c2 == 255 ? 255 : min(255, (c1 << 8) ~/ (255 - c2));
    }

    return Color.fromRGBO(
      colorDodgeChannel(baseColor.red, blendColor.red),
      colorDodgeChannel(baseColor.green, blendColor.green),
      colorDodgeChannel(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendColorBurn(Color baseColor, Color blendColor) {
    int colorBurnChannel(int c1, int c2) {
      return c2 == 0 ? 0 : max(0, 255 - ((255 - c1) << 8) ~/ c2);
    }

    return Color.fromRGBO(
      colorBurnChannel(baseColor.red, blendColor.red),
      colorBurnChannel(baseColor.green, blendColor.green),
      colorBurnChannel(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendHardLight(Color baseColor, Color blendColor) {
    int hardLightChannel(int c1, int c2) {
      return c2 < 128 ? 2 * c1 * c2 ~/ 255 : 255 - 2 * (255 - c1) * (255 - c2) ~/ 255;
    }

    return Color.fromRGBO(
      hardLightChannel(baseColor.red, blendColor.red),
      hardLightChannel(baseColor.green, blendColor.green),
      hardLightChannel(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendSoftLight(Color baseColor, Color blendColor) {
    num _softLightDodge(int c) {
      return c < 64 ? (16 * c - 12) * c + 4 * c : sqrt(c / 255) * 255;
    }

    int softLightChannel(int c1, int c2) {
      return c2 < 128 ? c1 - (255 - 2 * c2) * c1 * (255 - c1) ~/ 255 ~/ 255 : c1 + (2 * c2 - 255) * (c1 < 128 ? _softLightDodge(c1) : sqrt(c1 / 255)) ~/ 255;
    }

    return Color.fromRGBO(
      softLightChannel(baseColor.red, blendColor.red),
      softLightChannel(baseColor.green, blendColor.green),
      softLightChannel(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendDifference(Color baseColor, Color blendColor) {
    return Color.fromRGBO(
      (baseColor.red - blendColor.red).abs(),
      (baseColor.green - blendColor.green).abs(),
      (baseColor.blue - blendColor.blue).abs(),
      1,
    );
  }

  static Color blendExclusion(Color baseColor, Color blendColor) {
    int exclusionChannel(int baseC, int blendC) {
      return baseC + blendC - 2 * baseC * blendC ~/ 255;
    }

    return Color.fromRGBO(
      exclusionChannel(baseColor.red, blendColor.red),
      exclusionChannel(baseColor.green, blendColor.green),
      exclusionChannel(baseColor.blue, blendColor.blue),
      1,
    );
  }

  static Color blendHue(Color baseColor, Color blendColor) {
    final baseHSL = rgbToHsl(baseColor);
    final blendHSL = rgbToHsl(blendColor);

    return hslToRgb(blendHSL[0], baseHSL[1], baseHSL[2]);
  }

  static Color blendSaturation(Color baseColor, Color blendColor) {
    final baseHSL = rgbToHsl(baseColor);
    final blendHSL = rgbToHsl(blendColor);

    return hslToRgb(baseHSL[0], blendHSL[1], baseHSL[2]);
  }

  static Color blendTheColor(Color baseColor, Color blendColor) {
    final baseHSL = rgbToHsl(baseColor);
    final blendHSL = rgbToHsl(blendColor);

    return hslToRgb(blendHSL[0], blendHSL[1], baseHSL[2]);
  }

  static Color blendLuminosity(Color baseColor, Color blendColor) {
    final baseHSL = rgbToHsl(baseColor);
    final blendHSL = rgbToHsl(blendColor);

    return hslToRgb(baseHSL[0], baseHSL[1], blendHSL[2]);
  }

  static List<double> rgbToHsl(Color color) {
    final r = color.red / 255.0;
    final g = color.green / 255.0;
    final b = color.blue / 255.0;
    final maxColor = max(r, max(g, b));
    final minColor = min(r, min(g, b));
    double h = 0.0, s;
    final l = (maxColor + minColor) / 2.0;

    if (maxColor == minColor) {
      h = s = 0.0;
    } else {
      final d = maxColor - minColor;
      s = l > 0.5 ? d / (2.0 - maxColor - minColor) : d / (maxColor + minColor);
      if (maxColor == r) {
        h = (g - b) / d + (g < b ? 6.0 : 0.0);
      } else if (maxColor == g) {
        h = (b - r) / d + 2.0;
      } else if (maxColor == b) {
        h = (r - g) / d + 4.0;
      }
      h /= 6.0;
    }

    return [h, s, l];
  }

  static Color hslToRgb(double h, double s, double l) {
    double r, g, b;

    if (s == 0.0) {
      r = g = b = l;
    } else {
      final q = l < 0.5 ? l * (1.0 + s) : l + s - l * s;
      final p = 2.0 * l - q;
      r = _hue2rgb(p, q, h + 1.0 / 3.0);
      g = _hue2rgb(p, q, h);
      b = _hue2rgb(p, q, h - 1.0 / 3.0);
    }

    return Color.fromRGBO((r * 255).round(), (g * 255).round(), (b * 255).round(), 1);
  }

  static double _hue2rgb(double p, double q, double t) {
    if (t < 0.0) t += 1.0;
    if (t > 1.0) t -= 1.0;
    if (t < 1.0 / 6.0) return p + (q - p) * 6.0 * t;
    if (t < 1.0 / 2.0) return q;
    if (t < 2.0 / 3.0) return p + (q - p) * (2.0 / 3.0 - t) * 6.0;

    return p;
  }
}

class CachedImageInfo {
  final Color color;
  final int widthPixels;
  final int heightPixels;

  CachedImageInfo(this.color, this.widthPixels, this.heightPixels);
}
